/*
Copyright (c) 2018 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include "x11.h"
#include "font.h"
#include "ptyutil.h"
#include "util.h"
#include "matrix.h"
#include "unisel.h"

static int fo_normal_size= 20;
static int   wd_cols= 80;
static int   wd_rows= 24;
static int   dpix, dpiy;
static int  gfx_fd;
static int   exit_event_loop;
static int   click_x=-1, click_y=-1, click_at;
static SEL  *selector;
static int   imode;

#ifdef MUTERM_LOG
static FILE *log;
#endif

static int active_tab=-1;

static void deactivate(int tn);
static void activate(int tn);
static void start_tab(int tn);

typedef struct 
{
  int bg, fg, og;
  unsigned int normal[256], bold[256];
} color_t;

static color_t colors[9]=
 {
   { 0x000000, 0xaaaaaa, 0xffffff, },
   { 0x00004b, 0xaaaaaa, 0xffffff, },
   { 0x83713e, 0x000000, 0xffffff, },
   { 0x380000, 0xffffff, 0xffff00, },
   { 0x103164, 0xffffff, 0x00ffff, },
   { 0x004000, 0xffffff, 0xffff00, },
   { 0x404040, 0xffffff, 0xff0000, },
   { 0xcccccc, 0x000000, 0x8c4422, },
   { 0x00ffff, 0x000000, 0x000000, }
 };
static int curcolor;


static screen_t *SCR;

static char **child_argv;

typedef struct
{
  int present;
  int active;
  char *str;
  screen_t *screen;
  int tty_fd;
  pid_t child_pid;
} tab_t;

static tab_t tabs[9];


static void write_tty_str(char *b)
{
  if (!b) return ;
  int R;
  if (active_tab<0) return ;
  while(b[0])
  {
    R= write(tabs[active_tab].tty_fd, b, strlen(b));
    if (R<0) break;
    b+= R;
  }
}

void have_selection(uint32_t *wstr)
{
  if (!wstr || !wstr[0])
  {
     selection_lost();
     gfx_unset_selection();
  }
  else
  {
     char *str;
     int i;
     size_t wlen,slen;
     wlen= wstrlen(wstr);
     str= malloc(wlen*6+1);
     utf8_encode_string(wstr, wlen, (uint8_t*) str, &slen);
     str[slen]= 0;
     gfx_set_selection(str);
     for(i=0;i<8;i++)
       if (tabs[i].present && tabs[i].screen!=SCR)
          screen_select_none(tabs[i].screen);
       
  }
  if (wstr) free(wstr);
}

#ifdef MUDEBUG
void debug_trap()
{
  printf("DEBUG TRAP\n");
}
#endif

void key_event(int key, unsigned int wide, char *str)
{
  char *sel;
  char *b;
  char bu[2];
  int lapa;
  int i;
  if (imode) { help_key(key,wide,str); return ; }
  lapa= 1;
  b= NULL;
  if (key & GFX_K_CONTROL)
  {
    int z;
    z= key & ~GFX_K_CONTROL;
    if (z>='a' &&  z<='z') 
    {
      bu[0]= z-'a'+1;
      bu[1]= 0;
      b= bu;
      goto done;
    }
    if (z>='1' && z<='8')
    {
      int k;
      k=z-'1';
      if (!tabs[k].present) start_tab(k);
      activate(k);
      return ;
    }
#ifdef MUDEBUG
    if (z=='9')
    {
      debug_trap();
      return ;
    }
#endif
  }
  switch(key)
  {
  case GFX_K_PRT: b= str; break;
  case GFX_K_RETURN: b= "\r"; break;
  case GFX_K_TAB: b="\t"; break;
  case GFX_K_BACKSPACE: b= screen_bs_mode(SCR) ? "\177":"\b"; break;
  case GFX_K_SHIFT | GFX_K_INSERT: sel= gfx_get_selection('P'); goto insel;
  case GFX_K_CONTROL | GFX_K_INSERT: sel= gfx_get_selection('C'); goto insel;
  case GFX_K_CONTROL | GFX_K_RIGHT:
    for(i=1;i<8;i++)
      if (tabs[(active_tab+i)%8].present)
       { activate((active_tab+i)%8); break; }
    break;
  case GFX_K_CONTROL | GFX_K_LEFT:
    for(i=1;i<8;i++)
      if (tabs[(active_tab-i+8)%8].present)
       { activate((active_tab-i+8)%8); break; }
    break;
  
  case GFX_K_CONTROL | ' ':  gfx_copy_selection(); break;

  case GFX_K_SHIFT | GFX_K_PAGEUP: lapa= 0; screen_page_prev(SCR); break;
  case GFX_K_SHIFT | GFX_K_PAGEDOWN: lapa= 0; screen_page_next(SCR); break;

  case GFX_K_CONTROL | '\\': b= "\034"; break;
  case GFX_K_UP:  b= "\033OA"; break;
  case GFX_K_DOWN: b= "\033OB"; break;
  case GFX_K_LEFT: b= "\033OD"; break;
  case GFX_K_RIGHT: b= "\033OC"; break;
  case GFX_K_PAGEUP: b= "\033[5~"; break;
  case GFX_K_PAGEDOWN: b= "\033[6~"; break;
  case GFX_K_DELETE: b= screen_del_mode(SCR) ? "\177" : "\033[3~"; break;
  case GFX_K_INSERT: b= "\033[2~"; break;
  case GFX_K_ESCAPE: b= "\033"; break;
  case GFX_K_HOME: b= "\033OH"; break;
  case GFX_K_END: b= "\033OF"; break;
  default: return ;
  }
done:
  write_tty_str(b);
  if (lapa) screen_page_last(SCR);
  return;
insel:
  write_tty_str(sel);
  free(sel);
  if (lapa) screen_page_last(SCR);
  return ;
}

extern unsigned char RSTAfont1, RENDfont1;

static int fontH,fontW,fontA;


static unsigned int combine_channel(unsigned int B, int v, unsigned int F)
{
  return  (v*F+(255-v)*B)/255;
}

static int combine_colors(unsigned int bg, int v, unsigned int fg)
{
  return (combine_channel( (bg>>16)&0xff, v, (fg>>16)&0xff) << 16)
       | (combine_channel( (bg>>8)&0xff, v, (fg>>8)&0xff) << 8)
       | (combine_channel( (bg)&0xff, v, (fg)&0xff) );
}

static void init_colors()
{
  int c;
  unsigned int i;
  for(c=0;c<9;c++)
  {
    for(i=0;i<256;i++)
    {
      colors[c].normal[i]= combine_colors( colors[c].bg, i, colors[c].fg );
      colors[c].bold[i]= combine_colors( colors[c].bg, i, colors[c].og );
    }
  }
}



static void init_fonts()
{
  font_init(dpix);
  font_load_mem(0, &RSTAfont1, &RENDfont1-&RSTAfont1, fo_normal_size);

  int fw1, fh1, fa1;
  font_dimensions(0, &fw1,&fh1,&fa1);

  // fh1+= 8; fa1+= 8;

  font_set_cell(0,fw1,fh1,fa1);

  fontW= fw1; fontH= fh1; fontA= fa1;
}

static uint32_t **dbuffer, dW, dH;


static int cro, cco;
static int selb_row=-1, selb_col, sele_row, sele_col;

void set_cursor(int row, int col)
{
  cro= row; cco= col;
}

void set_sel(int br, int bc, int er, int ec)
{
  selb_row= br;
  selb_col= bc;
  sele_row= er;
  sele_col= ec;
}

void display_glyph(int row,int col, uint32_t C)
{
  unsigned int *comap;
  glyph_t *G;
  int ox,oy,x,y;
  unsigned char XO;
  uint32_t chr;

  ox= col*fontW;
  oy= (row+1)*fontH;
  if ((row==cro && col==cco) || (C&CELLATTR_REV)) XO= 255; else XO= 0;

  chr= C&CELL_CHRMASK;
  G= font_render(0,chr);

#define spos(r,c) ((r)*wd_cols+(c))
  if (selb_row!=-1 && spos(selb_row,selb_col)<=spos(row,col) &&
                      spos(sele_row,sele_col)>=spos(row,col))
    comap= colors[8].normal; 
  else if (C&CELLATTR_BOLD)
    comap= colors[curcolor].bold;
  else 
    comap= colors[curcolor].normal;
#undef spos

  for(y=0;y<fontH;y++)
    for(x=0;x<fontW;x++)
      dbuffer[y+oy][x+ox]= comap[G->data[x+y*fontW]^XO];
}

#if 0
static int matrix_dbg=0;

void dump(unsigned char *str,int len)
{
  int i;
  if (matrix_dbg) {
   printf("matrixdbg\n");
    for(i=0;i<len;i++) { if (i%16==0) printf("<\n"); 
       if (str[i]>=' ' && str[i]<127) printf("%c",str[i]); else printf(" %02x ", str[i]); }
     printf("<-----\n");
  }
}
#endif

static void tryexec(const char *fi)
{
  fprintf(stderr,"trying %s\n",fi);
  execl(fi, fi, NULL);
}

static void start_tab(int tn)
{
  if (tabs[tn].present) 
  {
    fprintf(stderr,"tab %d already present\n",tn);
    return ;
  }
  tabs[tn].tty_fd= ptymaster();
  if (tabs[tn].tty_fd<0) 
  {
    fprintf(stderr,"ptymaster() failed: %s\n",strerror(errno));
    return ;
  }
  tabs[tn].child_pid= fork();
  if (tabs[tn].child_pid==-1)
  {
    fprintf(stderr,"fork failed: %s\n",strerror(errno));
    close(tabs[tn].tty_fd);
    return ;
  }
  if (tabs[tn].child_pid) 
  {
    SELadd(selector, tabs[tn].tty_fd, tn+1, SELr);
    screen_set_fd(tabs[tn].screen, tabs[tn].tty_fd);
    tabs[tn].present= 1;
    return ;
  }

  if (ptyslave(tabs[tn].tty_fd))
  {
    fprintf(stderr,"ptyslave() failed: %s\n",strerror(errno));
    return ;
  }

  setenv("TERM","xterm",1);
  if (!child_argv[0]) 
  {
    fprintf(stderr,"bad usage: no command given to execute\n");
  }
  else 
  {
    if (execvp(child_argv[0], child_argv))
     fprintf(stderr,"can't execute %s: %s\n", child_argv[0], strerror(errno));
  }
  tryexec("/bin/bash");
  tryexec("/usr/bin/bash");
  tryexec("/bin/sh");
  exit(1);
}

static void init_selector()
{
  selector= SELnew();
  SELadd(selector, gfx_fd, 20, SELr);
}

static void draw_tab(int tn);

static void deactivate(int tn)
{
  int i;
  if (!tabs[tn].present) return ;
  /*  012345678 
          X            */
  tabs[tn].present= 0;
  tabs[tn].active= 0;
  tabs[tn].str[3]= '[';
  tabs[tn].str[5]= ']';
  screen_activate(tabs[tn].screen, 0);
  close(tabs[tn].tty_fd);
  tabs[tn].tty_fd= -1;
  waitpid(tabs[tn].child_pid,NULL,0);
  draw_tab(tn);
  if (tn==active_tab)
  {
    for(i=0;i<8;i++)
     if (tabs[(tn+i)%8].present) break;
    if (i==8) { exit_event_loop=1; return ;}
    activate((tn+i)%8);
  }
}

static void activate(int tn)
{
  tab_t *T;
  if (active_tab==tn) return ;
  if (!tabs[tn].screen) return ;
  if (!tabs[tn].present) return ;
  if (active_tab>=0)
  {
    T= tabs+active_tab;
    if (T->present)
    {
      T->str[3]= ' ';
      T->str[5]= ' ';
    }
    else
    {
      T->str[3]= '[';
      T->str[5]= ']';
    }
    draw_tab(active_tab);
  }
  active_tab= tn;
  T= tabs+active_tab;
  T->str[3]= '=';
  T->str[5]= '=';
  T->active= 1;
  draw_tab(active_tab);
  screen_activate(tabs[tn].screen,1);
  SCR= tabs[tn].screen;
  curcolor= tn;
  screen_redraw(SCR);
}

static void read_tty(int tn)
{
  unsigned char buf[80*24*6];
  int R;
  if (tn<0 || tn>=8) return ;

  if (tabs[tn].tty_fd<0)  goto remove;
  do {
    R= read(tabs[tn].tty_fd, buf, sizeof(buf));
    if (R<0) 
    {
      if (errno==EINTR) return;
      if (errno==EAGAIN) return;
#if 0
fprintf(log, "read: errno= %d %s\n", errno, strerror(errno));
fflush(log);
#endif
      goto remove;
    }
    if (R>0) screen_print_str(tabs[tn].screen,buf, R);
#ifdef MUTERM_LOG
    if (R>0) fwrite(buf, 1, R, log);
#endif
  } while(R>0);

  return ;
remove:
  SELrem(selector, -1, tn+1);
  deactivate(tn);
}

void display_modified()
{
  gfx_flush();
}

void event_loop()
{
  int ev;
  while(!exit_event_loop)
  {
    screen_updates(SCR);
    gfx_flush();
    for(ev=SELbeg(selector,-1);ev!=SELend;ev=SELnex(selector))
    {
      if ((ev&0xff)<=8 && (ev>>8)==SELr)
        read_tty( (ev & 0xff) -1 );
      else if (ev==SELEV(r,20)) gfx_event();
    }
  }
}


void mouse_event(int event,int button,int x,int y)
{
// these really belong in a header, yes?
#define MB_UP  4
#define MB_DOWN 5
#define MB_LEFT 1
#define MB_MID 2
#define MEV_PRESS 'P'
#define MEV_RELEASE 'R'
#define MEV_MOVE 'M'
#define MEV(X,Y) ((X)<<8 | (Y))
  int row;
  if (imode) return ;
  switch(MEV(event,button))
  {
    case MEV(MEV_PRESS,MB_UP):
      if (screen_mode(SCR))
        write_tty_str("\033[5~");
      else
        screen_page_prev(SCR);
      break;
    case MEV(MEV_PRESS,MB_DOWN):
      if (screen_mode(SCR))
        write_tty_str("\033[6~");
      else
	screen_page_next(SCR); 
      break;
    case MEV(MEV_PRESS,MB_LEFT):
       row= y/fontH-1;
       click_x= x; click_y= y;
       if (row<0)
       {
          click_at= 'B'; 
          return ;
       }
       click_at= 'A';
       gfx_pointer_move_events(1);
       screen_ppress(SCR, click_y/fontH-1, click_x/fontW);
       break;
    case MEV(MEV_RELEASE,MB_LEFT):
       gfx_pointer_move_events(0);
       if (click_at=='B')
       {
         int T;
         T= click_x/fontW/9;
         if (T<8)
         {
           if (!tabs[T].present) start_tab(T);
           activate(T);
         }
         else
         {
           imode= 'h';
           help_show();
           imode= 0;
           screen_redraw(SCR);
         }
         break;
       }
       screen_prelease(SCR);
       break;
    case MEV(MEV_MOVE,0):
       if (click_at!='A') break;
       screen_pmove(SCR, y/fontH-1, x/fontW);
       break;
    case MEV(MEV_PRESS, MB_MID):
       write_tty_str(gfx_get_selection('P'));
       break;
  }
#undef MB_UP
#undef MB_DOWN
#undef MB_LEFT
#undef MB_MID
#undef MEV_PRESS
#undef MEV_RELEASE
#undef MEV_MOVE
#undef MEV
}

void update_area(int col,int row, int wid, int hei)
{
  gfx_update(col*fontW, (row+1)*fontH, wid*fontW, hei*fontH);
}





void set_title(uint8_t *S)
{
  gfx_set_window_title(S);
}

void quit_request()
{
  exit_event_loop= 1;
}

void selection_lost()
{
  int i;
  for(i=0;i<8;i++)
    if (tabs[i].present)
      screen_select_none(tabs[i].screen);
}

static void init_tabs()
{
  int i;
  char buf[10];
  for(i=0;i<8;i++)
  {
    memset(buf,' ', 10);
    buf[9]= 0;
    buf[3]= '[';
    buf[4]= i+'1';
    buf[5]= ']';
    tabs[i].str= strdup(buf);
    tabs[i].tty_fd= -1;
    tabs[i].screen= screen_new(wd_rows, wd_cols);
  }
  tabs[8].str= strdup("  MENU ");
}

static void draw_tab(int tn)
{
  int j,k;
  int cc;
  uint32_t C;

  cc= curcolor;
  curcolor= tn;
  k= tn*9;
  for(j=0;tabs[tn].str[j];j++,k++)
  {
      C= tabs[tn].str[j];
      display_glyph(-1, k, C);
  }
  gfx_update(tn*9*fontW,0,9*fontW,fontH);
  curcolor= cc;
}

static void draw_bar()
{
  int i,j,k;
  int cc;
  uint32_t C;

  cc= curcolor;
  k= 0;
  C= 0;
  for(i=0;i<9;i++)
  {
    curcolor= i%8;
    for(j=0;tabs[i].str[j];j++,k++)
    {
      C= tabs[i].str[j];
      display_glyph(-1, k, C);
    }
  }
  gfx_update(0,0,dW,fontH);
  curcolor= cc;
}

static void sigchld_handle(int dummy)
{
}

int main(int argc,char **argv)
{
  signal(SIGCHLD, sigchld_handle);
#ifdef MUTERM_LOG
  log= fopen("/tmp/mulog", "wb");
#endif
  if (getenv("MUTERM_PRINT_UESC")) matrix_debug(1);
  child_argv= argv+1;
  if (gfx_init(NULL)) return 1;
  gfx_dpi(&dpix, &dpiy);
  init_fonts();
  init_colors();
  init_tabs();
  gfx_start(fontW*wd_cols, fontH*(wd_rows+1), "muterm");
  gfx_fd= gfx_descriptor();
  init_selector();
  gfx_buffer(&dbuffer,&dW,&dH);
  draw_bar();
  start_tab(0);
  activate(0);
  event_loop();
  gfx_close();
#ifdef MUTERM_LOG
  fclose(log);
#endif
  return 0;
}
