/*
Copyright (c) 2018 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include "udec.h"

#define BIT31        (1u<<31u)

static const unsigned utf8_ranges[]=  
{
  0x00000000, 0x0000007F,
  0x00000000, 0x00000000,
  0x00000080, 0x000007FF,
  0x00000800, 0x0000FFFF,
  0x00010000, 0x001FFFFF,
  0x00200000, 0x03FFFFFF,
  0x04000000, 0x7FFFFFFF,
  0x00000000, 0x00000000,
  0x00000000, 0x00000000 
};

static int leading_bits(unsigned char X) 
{
  if ((X&0x80)==0) return 0;   // OK, so we don't have any leading bits.
  if ((X&0x40)==0) return 1;   // The above failed, so we have one leading bit,
                               // but not two.
  if ((X&0x20)==0) return 2;   // and so on..
  if ((X&0x10)==0) return 3;
  if ((X&0x08)==0) return 4;
  if ((X&0x04)==0) return 5;
  if ((X&0x02)==0) return 6;
  // I don't distinguish between 7 leading bits and 8.
  // they are both illegal bytes.
  return 7;
}



void utf8decoder_reset(utf8decoder_t *dec)
{
  dec->size= 0;
  dec->len= 0;
  dec->outlen= 0;
}

void utf8decoder_init(utf8decoder_t *dec,int erract) 
{
  dec->erract= erract;
  utf8decoder_reset(dec);
}

static int put(utf8decoder_t *dec, unsigned chr)
{
  dec->outbuf[dec->outlen++]= chr;
  return dec->outlen;
}

static int uerror0(utf8decoder_t *dec)
{
  int i;
  switch(dec->erract) 
  {
  case UTF8_DELETE: 
    i= 0;
    break;
  case UTF8_REPORT: 
    for(i=0;i<dec->len;i++) dec->outbuf[i]= BIT31 | dec->pbuf[i];
    break;
  default:          
    for(i=0;i<dec->len;i++) dec->outbuf[i]= dec->erract; 
    break;
  }
  dec->len= 0;
  dec->size= 0;
  dec->outlen= i;
  return i;
}

static int uerror1(utf8decoder_t *dec, unsigned chr)
{
  dec->pbuf[dec->len++]= chr;
  return uerror0(dec);
}

static int start_sequence(utf8decoder_t *dec, int lb, unsigned chr)
{
  unsigned int mask;
  dec->size= lb;
  dec->len= 1;
  dec->pbuf[0]= chr;
  /**************************************
   If we have 2 leading bits, then the byte looks like:
      110x xxxx
   8-2 is 6, 1<<6 is:
      0010 00000
   when we subtract 1, it becomes:
      0001 11111
   this is our mask.
   **************************************/
  mask= (1<<(8-lb)) - 1;
  dec->accu= mask & chr;
  return dec->outlen;
}

static int check_valid(utf8decoder_t *dec)
{
  unsigned int A;
  A= dec->accu;
  if ((A &0xF000)==0xD000)
  {
    switch(A)
    {
      case 0xD800:                // invalid symbols in an UTF-8 stream
      case 0xDB7F:                // they are markers of some kind, I don't
      case 0xDB80:                // know, ask Markus Kuhn
      case 0xDBFF:
      case 0xDC00:
      case 0xDF80:
      case 0xDFFF:
        return uerror0(dec);         
    }
  }
  if (A==0xFFFE || A==0xFFFF)
    return uerror0(dec);
  dec->len= 0;
  dec->size= 0;
  dec->outbuf[0]= A;
  return dec->outlen=1;
}

int utf8decoder_chr(utf8decoder_t *dec,unsigned char chr) 
{
  int lb;
  unsigned int lo;

  if (chr==0xFF || chr==0xFE)    // always invalid..
     return uerror1(dec, chr);

  lb= leading_bits(chr);
  if (lb==0)                     // ASCII character : bit7 unset
  {
    if (dec->size) uerror0(dec); // abort the sequence if any
    return put(dec,chr);
  }
  /****
  from now on, lb can only be 1 thru 6.
  7 and 8 are already handled by the 0xff 0xfe condition
  we also handled case 0 above.
  ****/
  if (!dec->size)                // this is the initial state. 
  {
    if (lb!=1)                   // only allow cases 2..6
       return start_sequence(dec,lb,chr); 
    return uerror1(dec,chr);
  }

  if (lb!=1)                     // cases 2 .. 6 are start sequences, 
  {                              // we report and start over
    uerror0(dec);           
    return start_sequence(dec,lb,chr); 
  }

  /*** this is a proper continuation byte, 10xx xxxx ***/

  dec->pbuf[dec->len++]= chr;
                                  // 3F= 0011 1111
  dec->accu = (dec->accu << 6) | (chr & 0x3F); 
  if (dec->len<dec->size) return dec->outlen=0;

  lo= utf8_ranges[2*dec->len];
  if (dec->accu<lo)               // overlong encoding: report the sequence
     return uerror0(dec);         // including the current char as an error.
  return check_valid(dec);
}

int utf8decoder_end(utf8decoder_t *dec)
{
  if (!dec->size) { dec->len= 0; return dec->outlen=0; }
  return uerror0(dec);
}

